
"""
Module implementing representational similarity analysis (RSA) at the sensor
level.

Kriegeskorte, N., Mur, M., & Bandettini, P. A. (2008). Representational
similarity analysis - connecting the branches of systems neuroscience.
Frontiers in Systems Neuroscience, 2(November), 4.
https://doi.org/10.3389/neuro.06.004.2008

Authors
-------
Marijn van Vliet <w.m.vanvliet@gmail.com>
"""
import numpy as np
from scipy.spatial import distance
import mne
from mne.utils import logger

from .dsm import _n_items_from_dsm, dsm_array
from .searchlight import searchlight
from .rsa import rsa_array


def rsa_evokeds(evokeds, dsm_model, noise_cov=None, spatial_radius=0.04,
                temporal_radius=0.1, evoked_dsm_metric='correlation',
                evoked_dsm_params=dict(), rsa_metric='spearman', y=None,
                n_folds=1, picks=None, tmin=None, tmax=None, n_jobs=1,
                verbose=False):
    """Perform RSA in a searchlight pattern on evokeds.

    The output is an Evoked object where the "signal" at each sensor is
    the RSA, computed using all surrounding sensors.

    Parameters
    ----------
    evokeds : list of :class:`mne.Evoked`
        The evoked brain activity for each item. If you have more than one
        Evoked object per item (i.e. repetitions), you can use the ``y``
        parameter to assign evokeds to items.
    dsm_model : ndarray, shape (n, n) | (n * (n - 1) // 2,) | list of ndarray
        The model DSM, see :func:`compute_dsm`. For efficiency, you can give it
        in condensed form, meaning only the upper triangle of the matrix as a
        vector. See :func:`scipy.spatial.distance.squareform`. To perform RSA
        against multiple models at the same time, supply a list of model DSMs.

        Use :func:`compute_dsm` to compute DSMs.
    noise_cov : :class:`mne.Covariance` | None
        When specified, the data will by normalized using the noise covariance.
        This is recommended in all cases, but a hard requirement when the data
        contains sensors of different types. Defaults to None.
    spatial_radius : float | None
        The spatial radius of the searchlight patch in meters. All sensors
        within this radius will belong to the searchlight patch. Set to None to
        only perform the searchlight over time, flattening across sensors.
        Defaults to 0.04.
    temporal_radius : float | None
        The temporal radius of the searchlight patch in seconds. Set to None to
        only perform the searchlight over sensors, flattening across time.
        Defaults to 0.1.
    evoked_dsm_metric : str
        The metric to use to compute the DSM for the evokeds. This can be any
        metric supported by the scipy.distance.pdist function. See also the
        ``evoked_dsm_params`` parameter to specify and additional parameter for
        the distance function. Defaults to 'correlation'.
    evoked_dsm_params : dict
        Extra arguments for the distance metric used to compute the DSMs.
        Refer to :mod:`scipy.spatial.distance` for a list of all other metrics
        and their arguments. Defaults to an empty dictionary.
    rsa_metric : str
        The RSA metric to use to compare the DSMs. Valid options are:

        * 'spearman' for Spearman's correlation (the default)
        * 'pearson' for Pearson's correlation
        * 'kendall-tau-a' for Kendall's Tau (alpha variant)
        * 'partial' for partial Pearson correlations
        * 'partial-spearman' for partial Spearman correlations
        * 'regression' for linear regression weights

        Defaults to 'spearman'.
    y : ndarray of int, shape (n_items,) | None
        For each Evoked, a number indicating the item to which it belongs.
        When ``None``, each Evoked is assumed to belong to a different item.
        Defaults to ``None``.
    n_folds : int | None
        Number of cross-validation folds to use when computing the distance
        metric. Folds are created based on the ``y`` parameter. Specify
        ``None`` to use the maximum number of folds possible, given the data.
        Defaults to 1 (no cross-validation).
    picks : str | list | slice | None
        Channels to include. Slices and lists of integers will be interpreted
        as channel indices. In lists, channel *type* strings (e.g., ``['meg',
        'eeg']``) will pick channels of those types, channel *name* strings
        (e.g., ``['MEG0111', 'MEG2623']`` will pick the given channels. Can
        also be the string values "all" to pick all channels, or "data" to pick
        data channels. ``None`` (default) will pick all MEG and EEG channels,
        excluding those maked as "bad".
    tmin : float | None
        When set, searchlight patches will only be generated from subsequent
        time points starting from this time point. This value is given in
        seconds. Defaults to ``None``, in which case patches are generated
        starting from the first time point.
    tmax : float | None
        When set, searchlight patches will only be generated up to and
        including this time point. This value is given in seconds. Defaults to
        ``None``, in which case patches are generated up to and including the
        last time point.
    n_jobs : int
        The number of processes (=number of CPU cores) to use. Specify -1 to
        use all available cores. Defaults to 1.
    verbose : bool
        Whether to display a progress bar. In order for this to work, you need
        the tqdm python module installed. Defaults to False.

    Returns
    -------
    rsa : Evoked | list of Evoked
        The correlation values for each searchlight patch. When spatial_radius
        is set to None, there will only be one virtual sensor. When
        temporal_radius is set to None, there will only be one time point. When
        multiple models have been supplied, a list will be returned containing
        the RSA results for each model.

    See Also
    --------
    compute_dsm
    """
    one_model = type(dsm_model) != list
    if one_model:
        dsm_model = [dsm_model]

    logger.info(f'Performing RSA between Evokeds and {len(dsm_model)} model '
                'DSM(s)')

    # Check for compatibility of the evokeds and the model features
    for dsm in dsm_model:
        n_items = _n_items_from_dsm(dsm)
        if len(evokeds) != n_items and y is None:
            raise ValueError(
                'The number of evokeds (%d) should be equal to the '
                'number of items in `dsm_model` (%d). Alternatively, use '
                'the `y` parameter to assign evokeds to items.'
                % (len(evokeds), n_items))
        if y is not None and np.unique(y) != n_items:
            raise ValueError(
                'The number of items in `dsm_model` (%d) does not match '
                'the number of items encoded in the `y` matrix (%d).'
                % (n_items, len(np.unique(y))))

    times = evokeds[0].times
    for evoked in evokeds:
        if np.any(evoked.times != times):
            raise ValueError('Not all evokeds have the same time points.')

    # Convert the temporal radius to samples
    if temporal_radius is not None:
        temporal_radius = round(evokeds[0].info['sfreq'] * temporal_radius)
        if temporal_radius < 1:
            raise ValueError('Temporal radius is less than one sample.')

    # Normalize with the noise cov
    if noise_cov is not None:
        logger.info('    Whitening data using noise covariance')
        diag = spatial_radius is not None
        evokeds = [mne.whiten_evoked(evoked, noise_cov, diag=diag)
                   for evoked in evokeds]

    # Compute the distances between the sensors
    locs = np.vstack([ch['loc'][:3] for ch in evokeds[0].info['chs']])
    dist = distance.squareform(distance.pdist(locs))

    picks = mne.io.pick._picks_to_idx(evokeds[0].info, picks, none='data')
    if len(picks) != len(set(picks)):
        raise ValueError("`picks` are not unique. Please remove duplicates.")
    sel_samples = _tmin_tmax_to_indices(evokeds[0].times, tmin, tmax)

    if spatial_radius is not None:
        logger.info(f'    Spatial radius: {spatial_radius} meters')
        logger.info(f'    Using {len(picks)} sensors')
    if temporal_radius is not None:
        logger.info(f'    Temporal radius: {temporal_radius} samples')
        logger.info(f'    Time inverval: {tmin}-{tmax} seconds')

    # Perform the RSA
    X = np.array([evoked.data for evoked in evokeds])
    patches = searchlight(X.shape, dist=dist, spatial_radius=spatial_radius,
                          temporal_radius=temporal_radius,
                          sel_series=picks, sel_samples=sel_samples)
    data = rsa_array(X, dsm_model, patches, data_dsm_metric=evoked_dsm_metric,
                     data_dsm_params=evoked_dsm_params, rsa_metric=rsa_metric,
                     y=y, n_folds=n_folds, n_jobs=n_jobs, verbose=verbose)

    # Pack the result in an Evoked object
    if spatial_radius is not None:
        info = mne.pick_info(evokeds[0].info, picks)
    else:
        info = mne.create_info(['rsa'], evokeds[0].info['sfreq'])
    tmin = _construct_tmin(evokeds[0].times, sel_samples, temporal_radius)

    if one_model:
        return mne.EvokedArray(data[:, :, 0], info, tmin, comment='RSA',
                               nave=len(evokeds))
    else:
        return [mne.EvokedArray(data[:, :, i], info, tmin, comment='RSA',
                                nave=len(evokeds))
                for i in range(data.shape[-1])]


def rsa_epochs(epochs, dsm_model, noise_cov=None, spatial_radius=0.04,
               temporal_radius=0.1, epochs_dsm_metric='correlation',
               epochs_dsm_params=dict(), rsa_metric='spearman', y=None,
               n_folds=1, picks=None, tmin=None, tmax=None, n_jobs=1,
               verbose=False):
    """Perform RSA in a searchlight pattern on epochs.

    The output is an Evoked object where the "signal" at each sensor is
    the RSA, computed using all surrounding sensors.

    Parameters
    ----------
    epochs : instance of mne.Epochs
        The brain activity during the epochs. The event codes are used to
        distinguish between items.
    dsm_model : ndarray, shape (n, n) | (n * (n - 1) // 2,) | list of ndarray
        The model DSM, see :func:`compute_dsm`. For efficiency, you can give it
        in condensed form, meaning only the upper triangle of the matrix as a
        vector. See :func:`scipy.spatial.distance.squareform`. To perform RSA
        against multiple models at the same time, supply a list of model DSMs.

        Use :func:`compute_dsm` to compute DSMs.
    noise_cov : mne.Covariance | None
        When specified, the data will by normalized using the noise covariance.
        This is recommended in all cases, but a hard requirement when the data
        contains sensors of different types. Defaults to None.
    spatial_radius : floats | None
        The spatial radius of the searchlight patch in meters. All sensors
        within this radius will belong to the searchlight patch. Set to None to
        only perform the searchlight over time, flattening across sensors.
        Defaults to 0.04.
    temporal_radius : float | None
        The temporal radius of the searchlight patch in seconds. Set to None to
        only perform the searchlight over sensors, flattening across time.
        Defaults to 0.1.
    epochs_dsm_metric : str
        The metric to use to compute the DSM for the epochs. This can be any
        metric supported by the scipy.distance.pdist function. See also the
        ``epochs_dsm_params`` parameter to specify and additional parameter for
        the distance function. Defaults to 'correlation'.
    epochs_dsm_params : dict
        Extra arguments for the distance metric used to compute the DSMs.
        Refer to :mod:`scipy.spatial.distance` for a list of all other metrics
        and their arguments. Defaults to an empty dictionary.
    rsa_metric : str
        The RSA metric to use to compare the DSMs. Valid options are:

        * 'spearman' for Spearman's correlation (the default)
        * 'pearson' for Pearson's correlation
        * 'kendall-tau-a' for Kendall's Tau (alpha variant)
        * 'partial' for partial Pearson correlations
        * 'partial-spearman' for partial Spearman correlations
        * 'regression' for linear regression weights

        Defaults to 'spearman'.
    y : ndarray of int, shape (n_items,) | None
        For each Epoch, a number indicating the item to which it belongs.
        When ``None``, the event codes are used to differentiate between items.
        Defaults to ``None``.
    n_folds : int | None
        Number of cross-validation folds to use when computing the distance
        metric. Folds are created based on the ``y`` parameter, or the event
        codes if ``y`` is not specified. Specify ``None`` to use the maximum
        number of folds possible, given the data.
        Defaults to 1 (no cross-validation).
    picks : str | list | slice | None
        Channels to include. Slices and lists of integers will be interpreted
        as channel indices. In lists, channel *type* strings (e.g., ``['meg',
        'eeg']``) will pick channels of those types, channel *name* strings
        (e.g., ``['MEG0111', 'MEG2623']`` will pick the given channels. Can
        also be the string values "all" to pick all channels, or "data" to pick
        data channels. ``None`` (default) will pick all MEG and EEG channels,
        excluding those maked as "bad".
    tmin : float | None
        When set, searchlight patches will only be generated from subsequent
        time points starting from this time point. This value is given in
        seconds. Defaults to ``None``, in which case patches are generated
        starting from the first time point.
    tmax : float | None
        When set, searchlight patches will only be generated up to and
        including this time point. This value is given in seconds. Defaults to
        ``None``, in which case patches are generated up to and including the
        last time point.
    n_jobs : int
        The number of processes (=number of CPU cores) to use. Specify -1 to
        use all available cores. Defaults to 1.
    verbose : bool
        Whether to display a progress bar. In order for this to work, you need
        the tqdm python module installed. Defaults to False.

    Returns
    -------
    rsa : Evoked | list of Evoked
        The correlation values for each searchlight patch. When spatial_radius
        is set to None, there will only be one virtual sensor. When
        temporal_radius is set to None, there will only be one time point. When
        multiple models have been supplied, a list will be returned containing
        the RSA results for each model.

    See Also
    --------
    compute_dsm
    """
    one_model = type(dsm_model) is np.ndarray
    if one_model:
        dsm_model = [dsm_model]

    logger.info(f'Performing RSA between Epochs and {len(dsm_model)} model '
                'DSM(s)')

    if y is None:
        y_source = 'Epoch object'
        y = epochs.events[:, 2]
    else:
        y_source = '`y` matrix'

    # Check for compatibility of the evokeds and the model features
    for dsm in dsm_model:
        n_items = _n_items_from_dsm(dsm)
        if len(np.unique(y)) != n_items:
            raise ValueError(
                'The number of items in `dsm_model` (%d) does not match '
                'the number of items encoded in the %s (%d).'
                % (n_items, y_source, len(np.unique(y))))

    # Convert the temporal radius to samples
    if temporal_radius is not None:
        temporal_radius = round(epochs.info['sfreq'] * temporal_radius)
        if temporal_radius < 1:
            raise ValueError('Temporal radius is less than one sample.')

    # # Normalize with the noise cov
    # if noise_cov is not None:
    #     diag = spatial_radius is not None
    #     evokeds = [mne.whiten_evoked(evoked, noise_cov, diag=diag)
    #                for evoked in evokeds]

    # Compute the distances between the sensors
    locs = np.vstack([ch['loc'][:3] for ch in epochs.info['chs']])
    dist = distance.squareform(distance.pdist(locs))

    picks = mne.io.pick._picks_to_idx(epochs.info, picks, none='data')
    if len(picks) != len(set(picks)):
        raise ValueError("`picks` are not unique. Please remove duplicates.")
    sel_samples = _tmin_tmax_to_indices(epochs.times, tmin, tmax)

    if spatial_radius is not None:
        logger.info(f'    Spatial radius: {spatial_radius} meters')
        logger.info(f'    Using {len(picks)} sensors')
    if temporal_radius is not None:
        logger.info(f'    Temporal radius: {temporal_radius} samples')
        logger.info(f'    Time inverval: {tmin}-{tmax} seconds')

    # Perform the RSA
    X = epochs.get_data()
    patches = searchlight(X.shape, dist=dist, spatial_radius=spatial_radius,
                          temporal_radius=temporal_radius,
                          sel_series=picks, sel_samples=sel_samples)
    data = rsa_array(X, dsm_model, patches, data_dsm_metric=epochs_dsm_metric,
                     data_dsm_params=epochs_dsm_params, rsa_metric=rsa_metric,
                     y=y, n_folds=n_folds, n_jobs=n_jobs, verbose=verbose)

    # Pack the result in an Evoked object
    if spatial_radius is not None:
        info = epochs.info
        info = mne.pick_info(info, picks)
    else:
        info = mne.create_info(['rsa'], epochs.info['sfreq'])
    tmin = _construct_tmin(epochs.times, sel_samples, temporal_radius)

    if one_model:
        return mne.EvokedArray(data, info, tmin, comment='RSA',
                               nave=len(np.unique(y)))
    else:
        return [mne.EvokedArray(data[:, :, i], info, tmin, comment='RSA',
                                nave=len(np.unique(y)))
                for i in range(data.shape[-1])]


def dsm_evokeds(evokeds, noise_cov=None, spatial_radius=0.04,
                temporal_radius=0.1, dist_metric='correlation',
                dist_params=dict(), y=None, n_folds=None, picks=None,
                tmin=None, tmax=None):
    """Generate DSMs in a searchlight pattern on evokeds.

    Parameters
    ----------
    evokeds : list of mne.Evoked
        The evoked brain activity for each item. If you have more than one
        Evoked object per item (i.e. repetitions), you can use the ``y``
        parameter to assign evokeds to items.
    noise_cov : mne.Covariance | None
        When specified, the data will by normalized using the noise covariance.
        This is recommended in all cases, but a hard requirement when the data
        contains sensors of different types. Defaults to None.
    spatial_radius : floats | None
        The spatial radius of the searchlight patch in meters. All sensors
        within this radius will belong to the searchlight patch. Set to None to
        only perform the searchlight over time, flattening across sensors.
        Defaults to 0.04.
    temporal_radius : float | None
        The temporal radius of the searchlight patch in seconds. Set to None to
        only perform the searchlight over sensors, flattening across time.
        Defaults to 0.1.
    dist_metric : str
        The metric to use to compute the DSM for the evokeds. This can be any
        metric supported by the scipy.distance.pdist function. See also the
        ``dist_params`` parameter to specify and additional parameter for the
        distance function. Defaults to 'correlation'.
    dist_params : dict
        Extra arguments for the distance metric used to compute the DSMs.
        Refer to :mod:`scipy.spatial.distance` for a list of all other metrics
        and their arguments. Defaults to an empty dictionary.
    y : ndarray of int, shape (n_items,) | None
        For each Evoked, a number indicating the item to which it belongs.
        When ``None``, each Evoked is assumed to belong to a different item.
        Defaults to ``None``.
    n_folds : int | None
        Number of cross-validation folds to use when computing the distance
        metric. Folds are created based on the ``y`` parameter, or the event
        codes if ``y`` is not specified. Specify -1 to use the maximum number
        of folds possible, given the data.
        Defaults to 1 (no cross-validation).
    picks : str | list | slice | None
        Channels to include. Slices and lists of integers will be interpreted
        as channel indices. In lists, channel *type* strings (e.g., ``['meg',
        'eeg']``) will pick channels of those types, channel *name* strings
        (e.g., ``['MEG0111', 'MEG2623']`` will pick the given channels. Can
        also be the string values "all" to pick all channels, or "data" to pick
        data channels. ``None`` (default) will pick all MEG and EEG channels,
        excluding those maked as "bad".
    tmin : float | None
        When set, searchlight patches will only be generated from subsequent
        time points starting from this time point. This value is given in
        seconds. Defaults to ``None``, in which case patches are generated
        starting from the first time point.
    tmax : float | None
        When set, searchlight patches will only be generated up to and
        including this time point. This value is given in seconds. Defaults to
        ``None``, in which case patches are generated up to and including the
        last time point.

    Yields
    ------
    dsm : ndarray, shape (n_items, n_items)
        A DSM for each searchlight patch.
    """
    times = evokeds[0].times
    for evoked in evokeds:
        if np.any(evoked.times != times):
            raise ValueError('Not all evokeds have the same time points.')

    # Convert the temporal radius to samples
    if temporal_radius is not None:
        temporal_radius = round(evokeds[0].info['sfreq'] * temporal_radius)
        if temporal_radius < 1:
            raise ValueError('Temporal radius is less than one sample.')

    # Normalize with the noise cov
    if noise_cov is not None:
        diag = spatial_radius is not None
        evokeds = [mne.whiten_evoked(evoked, noise_cov, diag=diag)
                   for evoked in evokeds]

    # Compute the distances between the sensors
    locs = np.vstack([ch['loc'][:3] for ch in evokeds[0].info['chs']])
    dist = distance.squareform(distance.pdist(locs))

    picks = mne.io.pick._picks_to_idx(evokeds[0].info, picks, none='data')
    if len(picks) != len(set(picks)):
        raise ValueError("`picks` are not unique. Please remove duplicates.")
    sel_samples = _tmin_tmax_to_indices(times, tmin, tmax)

    # Compute the DSMs
    X = np.array([evoked.data for evoked in evokeds])
    patches = searchlight(X.shape, dist=dist, spatial_radius=spatial_radius,
                          temporal_radius=temporal_radius,
                          sel_series=picks, sel_samples=sel_samples)
    yield from dsm_array(X, patches, dist_metric=dist_metric,
                         dist_params=dist_params, y=y, n_folds=n_folds)


def dsm_epochs(epochs, noise_cov=None, spatial_radius=0.04,
               temporal_radius=0.1, dist_metric='correlation',
               dist_params=dict(), y=None, n_folds=None, picks=None,
               tmin=None, tmax=None):
    """Generate DSMs in a searchlight pattern on epochs.

    Parameters
    ----------
    epochs : instance of mne.Epochs
        The brain activity during the epochs. The event codes are used to
        distinguish between items.
    noise_cov : mne.Covariance | None
        When specified, the data will by normalized using the noise covariance.
        This is recommended in all cases, but a hard requirement when the data
        contains sensors of different types. Defaults to None.
    spatial_radius : floats | None
        The spatial radius of the searchlight patch in meters. All sensors
        within this radius will belong to the searchlight patch. Set to None to
        only perform the searchlight over time, flattening across sensors.
        Defaults to 0.04.
    temporal_radius : float | None
        The temporal radius of the searchlight patch in seconds. Set to None to
        only perform the searchlight over sensors, flattening across time.
        Defaults to 0.1.
    dist_metric : str
        The metric to use to compute the DSM for the epochs. This can be any
        metric supported by the scipy.distance.pdist function. See also the
        ``epochs_dsm_params`` parameter to specify and additional parameter for
        the distance function. Defaults to 'correlation'.
    dist_params : dict
        Extra arguments for the distance metric used to compute the DSMs.
        Refer to :mod:`scipy.spatial.distance` for a list of all other metrics
        and their arguments. Defaults to an empty dictionary.
    y : ndarray of int, shape (n_items,) | None
        For each Epoch, a number indicating the item to which it belongs.
        When ``None``, the event codes are used to differentiate between items.
        Defaults to ``None``.
    n_folds : int | None
        Number of cross-validation folds to use when computing the distance
        metric. Folds are created based on the ``y`` parameter, or the event
        codes if ``y`` is not specified. Specify -1 to use the maximum number
        of folds possible, given the data.
        Defaults to 1 (no cross-validation).
    picks : str | list | slice | None
        Channels to include. Slices and lists of integers will be interpreted
        as channel indices. In lists, channel *type* strings (e.g., ``['meg',
        'eeg']``) will pick channels of those types, channel *name* strings
        (e.g., ``['MEG0111', 'MEG2623']`` will pick the given channels. Can
        also be the string values "all" to pick all channels, or "data" to pick
        data channels. ``None`` (default) will pick all MEG and EEG channels,
        excluding those maked as "bad".
    tmin : float | None
        When set, searchlight patches will only be generated from subsequent
        time points starting from this time point. This value is given in
        seconds. Defaults to ``None``, in which case patches are generated
        starting from the first time point.
    tmax : float | None
        When set, searchlight patches will only be generated up to and
        including this time point. This value is given in seconds. Defaults to
        ``None``, in which case patches are generated up to and including the
        last time point.

    Yields
    ------
    dsm : ndarray, shape (n_items, n_items)
        A DSM for each searchlight patch.
    """
    if y is None:
        y = epochs.events[:, 2]

    # Convert the temporal radius to samples
    if temporal_radius is not None:
        temporal_radius = round(epochs.info['sfreq'] * temporal_radius)
        if temporal_radius < 1:
            raise ValueError('Temporal radius is less than one sample.')

    # # Normalize with the noise cov
    # if noise_cov is not None:
    #     diag = spatial_radius is not None
    #     evokeds = [mne.whiten_evoked(evoked, noise_cov, diag=diag)
    #                for evoked in evokeds]

    # Compute the distances between the sensors
    locs = np.vstack([ch['loc'][:3] for ch in epochs.info['chs']])
    dist = distance.squareform(distance.pdist(locs))

    picks = mne.io.pick._picks_to_idx(epochs.info, picks, none='data')
    if len(picks) != len(set(picks)):
        raise ValueError("`picks` are not unique. Please remove duplicates.")
    sel_samples = _tmin_tmax_to_indices(epochs.times, tmin, tmax)

    # Compute the DSMs
    X = epochs.get_data()
    patches = searchlight(X.shape, dist=dist, spatial_radius=spatial_radius,
                          temporal_radius=temporal_radius,
                          sel_series=picks, sel_samples=sel_samples)
    yield from dsm_array(X, patches, dist_metric=dist_metric,
                         dist_params=dist_params, y=y, n_folds=n_folds)


def _tmin_tmax_to_indices(times, tmin, tmax):
    """Convert tmin tmax parameters to an array of sample indices."""
    if tmin is None and tmax is None:
        sel_times = None
    else:
        if tmin is None:
            tmin = times[0]
        if tmax is None:
            tmin = times[-1]
        assert times[0] <= tmin < tmax <= times[-1], 'Invalid time range'
        sel_times = np.arange(np.searchsorted(times, tmin),
                              np.searchsorted(times, tmax) + 1)
    return sel_times


def _construct_tmin(times, sel_times, temporal_radius):
    if sel_times is None and temporal_radius is None:
        return times[int(round(len(times) / 2))]
    elif sel_times is not None and temporal_radius is None:
        return times[int(round(np.mean(sel_times)))]
    elif sel_times is None and temporal_radius is not None:
        return times[temporal_radius]
    elif sel_times is not None and temporal_radius is not None:
        return times[max(temporal_radius, sel_times[0])]
    # above cases are exhaustive



# encoding: utf-8
"""
Module implementing representational similarity analysis (RSA) at the source
level.

Kriegeskorte, N., Mur, M., & Bandettini, P. A. (2008). Representational
similarity analysis - connecting the branches of systems neuroscience.
Frontiers in Systems Neuroscience, 2(November), 4.
https://doi.org/10.3389/neuro.06.004.2008

Authors
-------
Marijn van Vliet <marijn.vanvliet@aalto.fi>
Annika Hultén <annika.hulten@aalto.fi>
Ossi Lehtonen <ossi.lehtonen@aalto.fi>
"""

from warnings import warn
import numpy as np
import mne
from mne.utils import logger
from scipy.linalg import block_diag
import nibabel as nib

from .dsm import _n_items_from_dsm, dsm_array
from .rsa import rsa_array
from .searchlight import searchlight
from .sensor_level import _tmin_tmax_to_indices, _construct_tmin


def rsa_stcs(stcs, dsm_model, src, spatial_radius=0.04, temporal_radius=0.1,
             stc_dsm_metric='correlation', stc_dsm_params=dict(),
             rsa_metric='spearman', y=None, n_folds=1, sel_vertices=None,
             tmin=None, tmax=None, n_jobs=1, verbose=False):
    """Perform RSA in a searchlight pattern on MNE-Python source estimates.

    The output is a source estimate where the "signal" at each source point is
    the RSA, computed for a patch surrounding the source point. Source estimate
    objects can be either defined along a cortical surface or volumetric.

    Parameters
    ----------
    stcs : list of mne.SourceEstimate | list of mne.VolSourceEstimate
        For each item, a source estimate for the brain activity.
    dsm_model : ndarray, shape (n, n) | (n * (n - 1) // 2,) | list of ndarray
        The model DSM, see :func:`compute_dsm`. For efficiency, you can give it
        in condensed form, meaning only the upper triangle of the matrix as a
        vector. See :func:`scipy.spatial.distance.squareform`. To perform RSA
        against multiple models at the same time, supply a list of model DSMs.

        Use :func:`compute_dsm` to compute DSMs.
    src : instance of mne.SourceSpaces
        The source space used by the source estimates specified in the `stcs`
        parameter.
    spatial_radius : float | None
        The spatial radius of the searchlight patch in meters. All source
        points within this radius will belong to the searchlight patch. Set to
        None to only perform the searchlight over time, flattening across
        sensors. Defaults to 0.04.
    temporal_radius : float | None
        The temporal radius of the searchlight patch in seconds. Set to None to
        only perform the searchlight over sensors, flattening across time.
        Defaults to 0.1.
    stc_dsm_metric : str
        The metric to use to compute the DSM for the source estimates. This can
        be any metric supported by the scipy.distance.pdist function. See also
        the ``stc_dsm_params`` parameter to specify and additional parameter
        for the distance function. Defaults to 'correlation'.
    stc_dsm_params : dict
        Extra arguments for the distance metric used to compute the DSMs.
        Refer to :mod:`scipy.spatial.distance` for a list of all other metrics
        and their arguments. Defaults to an empty dictionary.
    rsa_metric : str
        The RSA metric to use to compare the DSMs. Valid options are:

        * 'spearman' for Spearman's correlation (the default)
        * 'pearson' for Pearson's correlation
        * 'kendall-tau-a' for Kendall's Tau (alpha variant)
        * 'partial' for partial Pearson correlations
        * 'partial-spearman' for partial Spearman correlations
        * 'regression' for linear regression weights

        Defaults to 'spearman'.
    y : ndarray of int, shape (n_items,) | None
        For each source estimate, a number indicating the item to which it
        belongs. When ``None``, each source estimate is assumed to belong to a
        different item. Defaults to ``None``.
    n_folds : int | None
        Number of folds to use when using cross-validation to compute the
        evoked DSM metric. Specify ``None``, to use the maximum number of folds
        possible, given the data.
        Defaults to 1 (no cross-validation).
    sel_vertices : list of int | None
        When set, searchlight patches will only be generated for the subset of
        vertices/voxels with the given indices. Defaults to ``None``, in which
        case patches for all vertices/voxels are generated.
    tmin : float | None
        When set, searchlight patches will only be generated from subsequent
        time points starting from this time point. This value is given in
        seconds. Defaults to ``None``, in which case patches are generated
        starting from the first time point.
    tmax : float | None
        When set, searchlight patches will only be generated up to and
        including this time point. This value is given in seconds. Defaults to
        ``None``, in which case patches are generated up to and including the
        last time point.
    n_jobs : int
        The number of processes (=number of CPU cores) to use. Specify -1 to
        use all available cores. Defaults to 1.
    verbose : bool
        Whether to display a progress bar. In order for this to work, you need
        the tqdm python module installed. Defaults to False.

    Returns
    -------
    stc : SourceEstimate | VolSourceEstimate | list of SourceEstimate | list of VolSourceEstimate
        The correlation values for each searchlight patch. When spatial_radius
        is set to None, there will only be one vertex. When temporal_radius is
        set to None, there will only be one time point. When multiple models
        have been supplied, a list will be returned containing the RSA results
        for each model.

    See Also
    --------
    compute_dsm
    """  # noqa E501
    # Check for compatibility of the source estimates and the model features
    one_model = type(dsm_model) is np.ndarray
    if one_model:
        dsm_model = [dsm_model]

    # Check for compatibility of the stcs and the model features
    for dsm in dsm_model:
        n_items = _n_items_from_dsm(dsm)
        if len(stcs) != n_items and y is None:
            raise ValueError(
                'The number of source estimates (%d) should be equal to the '
                'number of items in `dsm_model` (%d). Alternatively, use '
                'the `y` parameter to assign source estimates to items.'
                % (len(stcs), n_items))
        if y is not None and len(np.unique(y)) != n_items:
            raise ValueError(
                'The number of items in `dsm_model` (%d) does not match '
                'the number of items encoded in the `y` matrix (%d).'
                % (n_items, len(np.unique(y))))

    _check_stcs_compatibility(stcs, src)
    dist = _get_distance_matrix(src, dist_lim=spatial_radius, n_jobs=n_jobs)

    if temporal_radius is not None:
        # Convert the temporal radius to samples
        temporal_radius = int(temporal_radius // stcs[0].tstep)

        if temporal_radius < 1:
            raise ValueError('Temporal radius is less than one sample.')

    sel_samples = _tmin_tmax_to_indices(stcs[0].times, tmin, tmax)

    # Perform the RSA
    X = np.array([stc.data for stc in stcs])
    patches = searchlight(X.shape, dist=dist, spatial_radius=spatial_radius,
                          temporal_radius=temporal_radius,
                          sel_series=sel_vertices, sel_samples=sel_samples)
    data = rsa_array(X, dsm_model, patches, data_dsm_metric=stc_dsm_metric,
                     data_dsm_params=stc_dsm_params, rsa_metric=rsa_metric,
                     y=y, n_folds=n_folds, n_jobs=n_jobs, verbose=verbose)

    # Pack the result in a SourceEstimate object
    if spatial_radius is not None:
        vertices = stcs[0].vertices
        if sel_vertices is not None:
            vertices = vertices[sel_vertices]
    else:
        if src.kind == 'volume':
            vertices = [np.array([1])]
        else:
            vertices = [np.array([1]), np.array([])]
    tmin = _construct_tmin(stcs[0].times, sel_samples, temporal_radius)
    tstep = stcs[0].tstep

    if one_model:
        if src.kind == 'volume':
            return mne.VolSourceEstimate(data, vertices, tmin, tstep,
                                         subject=stcs[0].subject)
        else:
            return mne.SourceEstimate(data, vertices, tmin, tstep,
                                      subject=stcs[0].subject)
    else:
        if src.kind == 'volume':
            return [mne.VolSourceEstimate(data[:, :, i], vertices, tmin, tstep,
                                          subject=stcs[0].subject)
                    for i in range(data.shape[-1])]
        else:
            return [mne.SourceEstimate(data[:, :, i], vertices, tmin, tstep,
                                       subject=stcs[0].subject)
                    for i in range(data.shape[-1])]


def dsm_stcs(stcs, src, spatial_radius=0.04, temporal_radius=0.1,
             dist_metric='sqeuclidean', dist_params=dict(), y=None,
             n_folds=None, sel_vertices=None, tmin=None, tmax=None, n_jobs=1,
             verbose=False):
    """Generate DSMs in a searchlight pattern on MNE-Python source estimates.

    DSMs are computed using a patch surrounding each source point. Source
    estimate objects can be either defined along a cortical surface or
    volumetric.

    Parameters
    ----------
    stcs : list of mne.SourceEstimate | list of mne.VolSourceEstimate
        For each item, a source estimate for the brain activity.
    src : instance of mne.SourceSpaces
        The source space used by the source estimates specified in the `stcs`
        parameter.
    spatial_radius : float | None
        The spatial radius of the searchlight patch in meters. All source
        points within this radius will belong to the searchlight patch. Set to
        None to only perform the searchlight over time, flattening across
        sensors. Defaults to 0.04.
    temporal_radius : float | None
        The temporal radius of the searchlight patch in seconds. Set to None to
        only perform the searchlight over sensors, flattening across time.
        Defaults to 0.1.
    dist_metric : str
        The metric to use to compute the DSM for the source estimates. This can
        be any metric supported by the scipy.distance.pdist function. See also
        the ``stc_dsm_params`` parameter to specify and additional parameter
        for the distance function. Defaults to 'sqeuclidean'.
    dist_params : dict
        Extra arguments for the distance metric used to compute the DSMs.
        Refer to :mod:`scipy.spatial.distance` for a list of all other metrics
        and their arguments. Defaults to an empty dictionary.
    y : ndarray of int, shape (n_items,) | None
        For each source estimate, a number indicating the item to which it
        belongs. When ``None``, each source estimate is assumed to belong to a
        different item. Defaults to ``None``.
    n_folds : int | None
        Number of folds to use when using cross-validation to compute the
        evoked DSM metric.  Defaults to ``None``, which means the maximum
        number of folds possible, given the data.
    sel_vertices : list of int | None
        When set, searchlight patches will only be generated for the subset of
        vertices/voxels with the given indices. Defaults to ``None``, in which
        case patches for all vertices/voxels are generated.
    tmin : float | None
        When set, searchlight patches will only be generated from subsequent
        time points starting from this time point. This value is given in
        seconds. Defaults to ``None``, in which case patches are generated
        starting from the first time point.
    tmax : float | None
        When set, searchlight patches will only be generated up to and
        including this time point. This value is given in seconds. Defaults to
        ``None``, in which case patches are generated up to and including the
        last time point.
    n_jobs : int
        The number of processes (=number of CPU cores) to use for the
        source-to-source distance computation. Specify -1 to use all available
        cores. Defaults to 1.

    Yields
    ------
    dsm : ndarray, shape (n_items, n_items)
        A DSM for each searchlight patch.
    """
    _check_stcs_compatibility(stcs, src)
    dist = _get_distance_matrix(src, dist_lim=spatial_radius, n_jobs=n_jobs)

    # Convert the temporal radius to samples
    if temporal_radius is not None:
        temporal_radius = int(temporal_radius // stcs[0].tstep)

        if temporal_radius < 1:
            raise ValueError('Temporal radius is less than one sample.')

    sel_samples = _tmin_tmax_to_indices(stcs[0].times, tmin, tmax)

    X = np.array([stc.data for stc in stcs])
    patches = searchlight(X.shape, dist=dist, spatial_radius=spatial_radius,
                          temporal_radius=temporal_radius,
                          sel_series=sel_vertices, sel_samples=sel_samples)
    yield from dsm_array(X, patches, dist_metric=dist_metric,
                         dist_params=dist_params, y=y, n_folds=n_folds)


def rsa_nifti(image, dsm_model, spatial_radius=0.01,
              image_dsm_metric='correlation', image_dsm_params=dict(),
              rsa_metric='spearman', y=None, n_folds=1, roi_mask=None,
              brain_mask=None, n_jobs=1, verbose=False):
    """Perform RSA in a searchlight pattern on Nibabel Nifti-like images.

    The output is a 3D Nifti image where the data at each voxel is is
    the RSA, computed for a patch surrounding the voxel. 

    .. versionadded:: 0.4

    Parameters
    ----------
    image : 4D Nifti-like image
        The Nitfi image data. The 4th dimension must contain the images
        for each item.
    dsm_model : ndarray, shape (n, n) | (n * (n - 1) // 2,) | list of ndarray
        The model DSM, see :func:`compute_dsm`. For efficiency, you can give it
        in condensed form, meaning only the upper triangle of the matrix as a
        vector. See :func:`scipy.spatial.distance.squareform`. To perform RSA
        against multiple models at the same time, supply a list of model DSMs.

        Use :func:`compute_dsm` to compute DSMs.
    spatial_radius : float
        The spatial radius of the searchlight patch in meters. All source
        points within this radius will belong to the searchlight patch.
        Defaults to 0.01.
    image_dsm_metric : str
        The metric to use to compute the DSM for the data. This can be
        any metric supported by the scipy.distance.pdist function. See also the
        ``image_dsm_params`` parameter to specify and additional parameter for
        the distance function. Defaults to 'correlation'.
    image_dsm_params : dict
        Extra arguments for the distance metric used to compute the DSMs.
        Refer to :mod:`scipy.spatial.distance` for a list of all other metrics
        and their arguments. Defaults to an empty dictionary.
    rsa_metric : str
        The RSA metric to use to compare the DSMs. Valid options are:

        * 'spearman' for Spearman's correlation (the default)
        * 'pearson' for Pearson's correlation
        * 'kendall-tau-a' for Kendall's Tau (alpha variant)
        * 'partial' for partial Pearson correlations
        * 'partial-spearman' for partial Spearman correlations
        * 'regression' for linear regression weights

        Defaults to 'spearman'.
    y : ndarray of int, shape (n_items,) | None
        For each source estimate, a number indicating the item to which it
        belongs. When ``None``, each source estimate is assumed to belong to a
        different item. Defaults to ``None``.
    n_folds : int | None
        Number of folds to use when using cross-validation to compute the
        evoked DSM metric. Specify ``None``, to use the maximum number of folds
        possible, given the data.
        Defaults to 1 (no cross-validation).
    roi_mask : 3D Nifti-like image | None
        When set, searchlight patches will only be generated for the subset of
        voxels with non-zero values in the given mask. This is useful for
        restricting the analysis to a region of interest (ROI). Note that while
        the center of the patches are all within the ROI, the patch itself may
        extend beyond the ROI boundaries.
        Defaults to ``None``, in which case patches for all voxels are
        generated.
    brain_mask : 3D Nifti-like image | None
        When set, searchlight patches are restricted to only contain voxels
        with non-zero values in the given mask. This is useful for make sure
        only information from inside the brain is used. In contrast to the
        `roi_mask`, searchlight patches will not use data outside of this mask.
        Defaults to ``None``, in which case all voxels are included in the
        analysis.
    n_jobs : int
        The number of processes (=number of CPU cores) to use. Specify -1 to
        use all available cores. Defaults to 1.
    verbose : bool
        Whether to display a progress bar. In order for this to work, you need
        the tqdm python module installed. Defaults to False.

    Returns
    -------
    rsa_results : 3D Nifti1Image | list of 3D Nifti1Image
        The correlation values for each searchlight patch. When multiple models
        have been supplied, a list will be returned containing the RSA results
        for each model.

    See Also
    --------
    compute_dsm
    """
    # Check for compatibility of the source estimates and the model features
    one_model = type(dsm_model) is np.ndarray
    if one_model:
        dsm_model = [dsm_model]

    if (not isinstance(image, tuple(nib.imageclasses.all_image_classes))
            or image.ndim != 4):
        raise ValueError('The image data must be 4-dimensional Nifti-like '
                         'images')

    # Check for compatibility of the BOLD images and the model features
    for dsm in dsm_model:
        n_items = _n_items_from_dsm(dsm)
        if image.shape[3] != n_items and y is None:
            raise ValueError(
                'The number of images (%d) should be equal to the '
                'number of items in `dsm_model` (%d). Alternatively, use '
                'the `y` parameter to assign evokeds to items.'
                % (image.shape[3], n_items))
        if y is not None and len(np.unique(y)) != n_items:
            raise ValueError(
                'The number of items in `dsm_model` (%d) does not match '
                'the number of items encoded in the `y` matrix (%d).'
                % (n_items, len(np.unique(y))))

    # Get data as (n_items x n_voxels)
    X = image.get_fdata().reshape(-1, image.shape[3]).T

    # Find voxel positions
    voxels = np.array(list(np.ndindex(image.shape[:-1])))
    voxel_loc = voxels @ image.affine[:3, :3]
    voxel_loc /= 1000  # convert position from mm to meters

    # Apply masks
    result_mask = np.ones(image.shape[:3], dtype=np.bool)
    if brain_mask is not None:
        if brain_mask.ndim != 3 or brain_mask.shape != image.shape[:3]:
            raise ValueError('Brain mask must be a 3-dimensional Nifi-like '
                             'image with the same dimensions as the data '
                             'image')
        brain_mask = brain_mask.get_fdata() != 0
        result_mask &= brain_mask
        brain_mask = brain_mask.ravel()
        X = X[:, brain_mask]
        voxel_loc = voxel_loc[brain_mask]
    if roi_mask is not None:
        if roi_mask.ndim != 3 or roi_mask.shape != image.shape[:3]:
            raise ValueError('ROI mask must be a 3-dimensional Nifi-like '
                             'image with the same dimensions as the data '
                             'image')
        roi_mask = roi_mask.get_fdata() != 0
        result_mask &= roi_mask
        roi_mask = roi_mask.ravel()
        if brain_mask is not None:
            roi_mask = roi_mask[brain_mask]
        roi_mask = np.flatnonzero(roi_mask)

    # Compute distances between voxels
    logger.info('Computing distances...')
    from sklearn.neighbors import NearestNeighbors
    nn = NearestNeighbors(radius=spatial_radius, n_jobs=n_jobs).fit(voxel_loc)
    dist = nn.radius_neighbors_graph(mode='distance')

    # Perform the RSA
    patches = searchlight(X.shape, dist=dist, spatial_radius=spatial_radius,
                          temporal_radius=None, sel_series=roi_mask)
    rsa_result = rsa_array(X, dsm_model, patches,
                           data_dsm_metric=image_dsm_metric,
                           data_dsm_params=image_dsm_params,
                           rsa_metric=rsa_metric, y=y, n_folds=n_folds,
                           n_jobs=n_jobs, verbose=verbose)

    if one_model:
        data = np.zeros(image.shape[:3])
        data[result_mask] = rsa_result
        return nib.Nifti1Image(data, image.affine, image.header)
    else:
        results = []
        for i in range(rsa_result.shape[-1]):
            data = np.zeros(image.shape[:3])
            data[result_mask] = rsa_result[:, i]
            results.append(nib.Nifti1Image(data, image.affine, image.header))
        return results


def dsm_nifti(image, spatial_radius=0.01, dist_metric='correlation',
              dist_params=dict(), y=None, n_folds=1, roi_mask=None,
              brain_mask=None, n_jobs=1, verbose=False):
    """Generate DSMs in a searchlight pattern on Nibabel Nifty-like images.

    DSMs are computed using a patch surrounding each voxel.

    .. versionadded:: 0.4

    Parameters
    ----------
    image : 4D Nifti-like image
        The Nitfi image data. The 4th dimension must contain the images
        for each item.
    spatial_radius : float
        The spatial radius of the searchlight patch in meters. All source
        points within this radius will belong to the searchlight patch.
        Defaults to 0.01.
    dist_metric : str
        The metric to use to compute the DSM for the data. This can be
        any metric supported by the scipy.distance.pdist function. See also the
        ``dist_params`` parameter to specify and additional parameter for
        the distance function. Defaults to 'correlation'.
    dist_params : dict
        Extra arguments for the distance metric used to compute the DSMs.
        Refer to :mod:`scipy.spatial.distance` for a list of all other metrics
        and their arguments. Defaults to an empty dictionary.
    y : ndarray of int, shape (n_items,) | None
        For each source estimate, a number indicating the item to which it
        belongs. When ``None``, each source estimate is assumed to belong to a
        different item. Defaults to ``None``.
    n_folds : int | None
        Number of folds to use when using cross-validation to compute the
        evoked DSM metric. Specify ``None``, to use the maximum number of folds
        possible, given the data.
        Defaults to 1 (no cross-validation).
    roi_mask : 3D Nifti-like image | None
        When set, searchlight patches will only be generated for the subset of
        voxels with non-zero values in the given mask. This is useful for
        restricting the analysis to a region of interest (ROI). Note that while
        the center of the patches are all within the ROI, the patch itself may
        extend beyond the ROI boundaries.
        Defaults to ``None``, in which case patches for all voxels are
        generated.
    brain_mask : 3D Nifti-like image | None
        When set, searchlight patches are restricted to only contain voxels
        with non-zero values in the given mask. This is useful for make sure
        only information from inside the brain is used. In contrast to the
        `roi_mask`, searchlight patches will not use data outside of this mask.
        Defaults to ``None``, in which case all voxels are included in the
        analysis.
    n_jobs : int
        The number of processes (=number of CPU cores) to use. Specify -1 to
        use all available cores. Defaults to 1.
    verbose : bool
        Whether to display a progress bar. In order for this to work, you need
        the tqdm python module installed. Defaults to False.

    Yields
    ------
    dsm : ndarray, shape (n_items, n_items)
        A DSM for each searchlight patch.
    """
    if (not isinstance(image, tuple(nib.imageclasses.all_image_classes))
            or image.ndim != 4):
        raise ValueError('The image data must be 4-dimensional Nifti-like '
                         'images')

    # Get data as (n_items x n_voxels)
    X = image.get_fdata().reshape(-1, image.shape[3]).T

    # Find voxel positions
    voxels = np.array(list(np.ndindex(image.shape[:-1])))
    voxel_loc = voxels @ image.affine[:3, :3]
    voxel_loc /= 1000  # convert position from mm to meters

    # Apply masks
    result_mask = np.ones(image.shape[:3], dtype=np.bool)
    if brain_mask is not None:
        if brain_mask.ndim != 3 or brain_mask.shape != image.shape[:3]:
            raise ValueError('Brain mask must be a 3-dimensional Nifi-like '
                             'image with the same dimensions as the data '
                             'image')
        brain_mask = brain_mask.get_fdata() != 0
        result_mask &= brain_mask
        brain_mask = brain_mask.ravel()
        X = X[:, brain_mask]
        voxel_loc = voxel_loc[brain_mask]
    if roi_mask is not None:
        if roi_mask.ndim != 3 or roi_mask.shape != image.shape[:3]:
            raise ValueError('ROI mask must be a 3-dimensional Nifi-like '
                             'image with the same dimensions as the data '
                             'image')
        roi_mask = roi_mask.get_fdata() != 0
        result_mask &= roi_mask
        roi_mask = roi_mask.ravel()
        if brain_mask is not None:
            roi_mask = roi_mask[brain_mask]
        roi_mask = np.flatnonzero(roi_mask)

    # Compute distances between voxels
    logger.info('Computing distances...')
    from sklearn.neighbors import NearestNeighbors
    nn = NearestNeighbors(radius=spatial_radius, n_jobs=n_jobs).fit(voxel_loc)
    dist = nn.radius_neighbors_graph(mode='distance')

    # Compute DSMs
    patches = searchlight(X.shape, dist=dist, spatial_radius=spatial_radius,
                          temporal_radius=None, sel_series=roi_mask)
    yield from dsm_array(X, patches, dist_metric=dist_metric,
                         dist_params=dist_params, y=y, n_folds=n_folds,
                         n_jobs=n_jobs, verbose=verbose)


def _check_stcs_compatibility(stcs, src):
    """Check for compatibility of the source estimates and source space."""
    if src.kind == 'volume' and not isinstance(stcs[0], mne.VolSourceEstimate):
        raise ValueError(f'Volume source estimates provided, but not a volume '
                         f'source space (src.kind={src.kind}).')

    for stc in stcs:
        if src.kind == 'volume':
            if np.any(stc.vertices != src[0]['vertno']):
                raise ValueError('Not all source estimates have the same '
                                 'vertices.')
        else:
            for src_hemi, stc_hemi_vertno in zip(src, stcs[0].vertices):
                if np.any(stc_hemi_vertno != src_hemi['vertno']):
                    raise ValueError('Not all source estimates have the same '
                                     'vertices.')

    times = stcs[0].times
    for stc in stcs:
        if np.any(stc.times != times):
            raise ValueError('Not all source estimates have the same '
                             'time points.')


def _get_distance_matrix(src, dist_lim, n_jobs=1):
    """Get vertex-to-vertex distance matrix from source space.

    During inverse computation, the source space was downsampled (i.e. using
    ico4). Construct vertex-to-vertex distance matrices using only the
    vertices that are defined in the source solution.

    Parameters
    ----------
    src : mne.SourceSpaces
        The source space to get the distance matrix for.
    dist_lim : float
        Maximum distance required. We don't care about distances beyond this
        maximum.
    n_jobs : int
        Number of CPU cores to use if distance computation is necessary.
        Defaults to 1.

    Returns
    -------
    dist : ndarray (n_vertices, n_vertices)
        The vertex-to-vertex distance matrix.
    """
    dist = []

    # Check if distances have been pre-computed in the given source space. Give
    # a warning if the pre-computed distances may have had a too limited
    # dist_lim setting.
    needs_distance_computation = False
    for hemi in src:
        if 'dist' not in hemi or hemi['dist'] is None:
            needs_distance_computation = True
        else:
            if hemi['dist_limit'][0] < dist_lim:
                warn(f'Source space has pre-computed distances, but all '
                     f'distances are smaller than the searchlight radius '
                     f'({dist_lim}). You may want to consider recomputing '
                     f'the source space distances using the '
                     f'mne.add_source_space_distances function.')

    if needs_distance_computation:
        if dist_lim is None:
            dist_lim = np.inf
        if src.kind == 'volume':
            src = _add_volume_source_space_distances(src, dist_lim)
        else:
            src = mne.add_source_space_distances(src, dist_lim, n_jobs=n_jobs)

    for hemi in src:
        inuse = np.flatnonzero(hemi['inuse'])
        dist.append(hemi['dist'][np.ix_(inuse, inuse)].toarray())

    # Collect the distances in a single matrix
    dist = block_diag(*dist)
    dist[dist == 0] = np.inf  # Across hemisphere distance is infinity
    dist.flat[::dist.shape[0] + 1] = 0  # Distance to yourself is zero

    return dist


def _add_volume_source_space_distances(src, dist_limit):
    """Compute the distance between voxels in a volume source space.

    Operates in-place!

    Code is mostly taken from `mne.add_source_space_distances`.

    Parameters
    ----------
    src : instance of mne.SourceSpaces
        The volume source space to compute the voxel-wise distances for.
    dist_limit : float
        The maximum distance (in meters) to consider. Voxels that are further
        apart than this distance will have a distance of infinity. Use this to
        reduce computation time.

    Returns
    -------
    src : instance of mne.SourceSpaces
        The volume source space, now with the 'dist' and 'dist_limit' fields
        set.
    """
    # Lazy import to not have to load the huge scipy module every time mne_rsa
    # get's loaded.
    from scipy.sparse import csr_matrix
    assert src.kind == 'volume'
    n_sources = src[0]['np']
    neighbors = np.array(src[0]['neighbor_vert'])
    rows, cols = np.nonzero(neighbors != -1)
    cols = neighbors[(rows, cols)]
    dist = np.linalg.norm(src[0]['rr'][rows, :] - src[0]['rr'][cols, :],
                          axis=1)
    con_matrix = csr_matrix((dist, (rows, cols)), shape=(n_sources, n_sources))
    dist = mne.source_space._do_src_distances(con_matrix, src[0]['vertno'],
                                              np.arange(src[0]['nuse']),
                                              dist_limit)[0]
    d = dist.ravel()  # already float32
    idx = d > 0
    d = d[idx]
    i, j = np.meshgrid(src[0]['vertno'], src[0]['vertno'])
    i = i.ravel()[idx]
    j = j.ravel()[idx]
    src[0]['dist'] = csr_matrix((d, (i, j)), shape=(n_sources, n_sources))
    src[0]['dist_limit'] = np.array([dist_limit], np.float32)
    return src


def make_mri_con_matrix(img):
    from scipy.sparse import csr_matrix
    # Create 3 x 3 x 3 cube of (ijk) indices, centered around (0, 0, 0)
    cube = np.array(list(np.ndindex(3, 3, 3))) - [1, 1, 1]
    # Remove center of the cube
    cube = np.delete(cube, len(cube) // 2, axis=0)
    # Compute distance from all points in the cube to the center
    dist = np.linalg.norm(cube @ img.affine[:3, :3], axis=1)
    # Copy the cube, centering it around each voxel
    voxels = np.array(list(np.ndindex(*img.shape[:3])))
    neighbours = voxels[:, :, np.newaxis] + cube.T[np.newaxis, :, :]
    assert neighbours.shape == (len(voxels), 3, len(cube))
    # Convert ijk coordinates to voxel numbers
    neighbours = np.ravel_multi_index(
        (neighbours[:, 0, :], neighbours[:, 1, :], neighbours[:, 2, :]),
        img.shape[:3],
        mode='clip',
    )
    rows = np.repeat(np.arange(neighbours.shape[0]), neighbours.shape[1])
    cols = neighbours.ravel()
    dist = np.tile(dist, neighbours.shape[0])
    con_matrix = csr_matrix((dist, (rows, cols)),
                            shape=(len(voxels), len(voxels)))
    return con_matrix



from mne.utils import logger
import numpy as np
from sklearn.model_selection import StratifiedKFold
from sklearn.preprocessing import OneHotEncoder


def _create_folds(X, y, n_folds=None):
    """Split the observations in X into stratified folds."""
    if y is None:
        # No folding
        return X[np.newaxis, ...]

    y = np.asarray(y)
    if len(y) != len(X):
        raise ValueError(f'The length of y ({len(y)}) does not match the '
                         f'number of items ({len(X)}).')

    y_one_hot = _convert_to_one_hot(y)
    n_items = y_one_hot.shape[1]

    if n_folds is None:
        # Set n_folds to maximum value
        n_folds = len(X) // n_items
        logger.info(f'Automatic dermination of folds: {n_folds}'
                    + ' (no cross-validation)' if n_folds == 1 else '')

    if n_folds == 1:
        # Making one fold is easy
        folds = [_compute_item_means(X, y_one_hot)]
    else:
        folds = []
        for _, fold in StratifiedKFold(n_folds).split(X, y):
            folds.append(_compute_item_means(X, y_one_hot, fold))
    return np.array(folds)


def _convert_to_one_hot(y):
    """Convert the labels in y to one-hot encoding."""
    y = np.asarray(y)
    if y.ndim == 1:
        y = y[:, np.newaxis]

    if y.ndim == 2 and y.shape[1] == 1:
        # y needs to be converted
        enc = OneHotEncoder(categories='auto').fit(y)
        return enc.transform(y).toarray()
    elif y.ndim > 2:
        raise ValueError('Wrong number of dimensions for `y`.')
    else:
        # y is probably already in one-hot form. We're not going to test this
        # explicitly, as it would take too long.
        return y


def _compute_item_means(X, y_one_hot, fold=slice(None)):
    """Compute the mean data for each item inside a fold."""
    X = X[fold]
    y_one_hot = y_one_hot[fold]
    n_per_class = y_one_hot.sum(axis=0)

    # The following computations go much faster when X is flattened.
    orig_shape = X.shape
    X_flat = X.reshape(len(X), -1)

    # Compute the mean for each item using matrix multiplication
    means = (y_one_hot.T @ X_flat) / n_per_class[:, np.newaxis]

    # Undo the flattening of X
    return means.reshape((len(means),) + orig_shape[1:])



import numpy as np
from mne.utils import logger


class searchlight:
    """Generate indices for searchlight patches.

    Generates a sequence of tuples that can be used to index a data array.
    Depending on the spatial and temporal radius, each tuple extracts a
    searchlight patch along time, space or both.

    This function is flexible in regards to shape of the data array. The
    intepretation of the dimensions is as follows::

    4 or more dimensions
        ``(n_folds, n_items, n_series, n_samples, ...)``

    3 dimensions
        ``(n_items, n_series, n_samples)``

    2 dimensions
        ``(n_items, n_series)`` when ``spatial_radius`` is not ``None``.
        ``(n_items, n_samples)`` when ``temporal_radius`` is not ``None``.

    1 dimension
        ``(n_items,)``

    The returned tuples will match the dimensions of the data array.

    Parameters
    ----------
    shape : tuple of int
        The shape of the data array to compute the searchlight patches for, as
        obtained with the ``.shape`` attribute.
    dist : ndarray or sparse matrix, shape (n_series, n_series) | None
        The distances between all source points or sensors in meters.
        This parameter needs to be specified if a ``spatial_radius`` is set.

        Since the distance matrix can be huge, sparse matrices are also
        supported. When the distance matrix is sparse, all zero distances are
        treated as infinity. This allows you to skip far away points during
        your distance computations.

        Defaults to ``None``.
    spatial_radius : floats | None
        The spatial radius of the searchlight patch in meters. All source
        points within this radius will belong to the searchlight patch. Set to
        None to only perform the searchlight over time. When this parameter is
        set, the ``dist`` parameter must also be specified. Defaults to
        ``None``.
    temporal_radius : float | None
        The temporal radius of the searchlight patch in samples. Set to
        ``None`` to only perform the searchlight over sensors/source points.
        Defaults to ``None``.
    sel_series : ndarray, shape (n_selected_series,) | None
        When set, searchlight patches will only be generated for the subset of
        time series with the given indices. Defaults to ``None``, in which case
        patches for all series are generated.
    sel_samples : ndarray, shape (n_selected_series,) | None
        When set, searchlight patches will only be generated for the subset of
        time samples with the given indices. Defaults to ``None``, in which
        case patches for all samples are generated.

    Yields
    ------
    patch : tuple of (slice | ndarray)
        A single searchlight patch. Each element of the tuple corresponds to a
        dimension of the data array and can be used to index along this
        dimension to extract the searchlight patch.

    Attributes
    ----------
    shape
    """
    def __init__(self, shape, dist=None, spatial_radius=None,
                 temporal_radius=None, sel_series=None,
                 sel_samples=None):
        # Interpret the dimensions of the data array (see docstring)
        n_dims = len(shape)
        if n_dims >= 4:
            self.series_dim = 2
            self.samples_dim = 3
        elif n_dims == 3:
            self.series_dim = 1
            self.samples_dim = 2
        elif n_dims == 2 and spatial_radius is not None:
            self.series_dim = 1
            self.samples_dim = None
        elif n_dims == 2 and temporal_radius is not None:
            self.series_dim = None
            self.samples_dim = 1
        else:
            self.series_dim = None
            self.samples_dim = None

        self.dist = dist
        self.spatial_radius = spatial_radius
        self.temporal_radius = temporal_radius
        self.sel_series = sel_series
        self.sel_samples = sel_samples

        # Will we be creating spatial searchlight patches?
        if self.spatial_radius is not None:
            if self.dist is None:
                raise ValueError('A spatial radius was requested, but no '
                                 'distance information was specified '
                                 '(=dist parameter).')
            if self.series_dim is None:
                raise ValueError('Cannot create spatial searchlight patches: '
                                 f'the provided data matrix {shape} has no '
                                 'spatial dimension.')
            if self.sel_series is None:
                self.sel_series = np.arange(shape[self.series_dim])

            # Compressed Sparse Row format is optimal for our computations
            from scipy.sparse import issparse
            if issparse(self.dist):
                self.dist = self.dist.tocsr()
        else:
            self.sel_series = None

        # Will we be creating temporal searchlight patches?
        if temporal_radius is not None:
            if self.samples_dim is None:
                raise ValueError('Cannot create temporal searchlight patches: '
                                 f'the provided data matrix {shape} has no '
                                 'temporal dimension.')
            n_samples = shape[self.samples_dim]
            if self.sel_samples is None:
                self.sel_samples = np.arange(n_samples)
            # Get a valid time range given the size of the sliding windows
            self.time_centers = [
                t for t in self.sel_samples
                if t - temporal_radius >= 0 and t + temporal_radius < n_samples
            ]
        else:
            self.sel_samples = None

        # Create a template for the patches that will be generated that is
        # compatible with the data array dimensions. By default, we select
        # everything along every dimension. This template will be filled-in
        # inside the __iter__ function.
        self.patch_template = [slice(None)] * n_dims

        # Setup the main generator function that will be providing the
        # searchlight patches.
        if (self.spatial_radius is not None
                and self.temporal_radius is not None):
            self._generator = self._iter_spatio_temporal()
        elif self.spatial_radius is not None:
            self._generator = self._iter_spatial()
        elif self.temporal_radius is not None:
            self._generator = self._iter_temporal()
        else:
            # Single searchlight patch only
            self._generator = iter([tuple(self.patch_template)])

    def __iter__(self):
        return self

    def __next__(self):
        """Generate searchlight patches."""
        return next(self._generator)

    def _iter_spatio_temporal(self):
        """Generate spatio-temporal searchlight patches."""
        logger.info('Creating spatio-temporal searchlight patches')
        patch = list(self.patch_template)  # Copy the template
        for series in self.sel_series:
            # Compute all spatial locations in the searchligh path.
            spat_ind = _get_in_radius(self.dist, series, self.spatial_radius)
            patch[self.series_dim] = spat_ind
            for sample in self.time_centers:
                temp_ind = slice(sample - self.temporal_radius,
                                 sample + self.temporal_radius + 1)
                patch[self.samples_dim] = temp_ind
                yield tuple(patch)

    def _iter_spatial(self):
        """Generate spatial searchlight patches only."""
        logger.info('Creating spatial searchlight patches')
        patch = list(self.patch_template)  # Copy the template
        for series in self.sel_series:
            spat_ind = _get_in_radius(self.dist, series, self.spatial_radius)
            patch[self.series_dim] = spat_ind
            yield tuple(patch)

    def _iter_temporal(self):
        """Generate temporal searchlight patches only."""
        logger.info('Creating temporal searchlight patches')
        patch = list(self.patch_template)  # Copy the template
        for sample in self.time_centers:
            patch[self.samples_dim] = slice(sample - self.temporal_radius,
                                            sample + self.temporal_radius + 1)
            yield tuple(patch)

    @property
    def shape(self):
        """Number of generated patches along multiple dimensions.

        This is useful for re-shaping the result obtained after consuming the
        this generator.

        Returns
        -------
        shape : tuple of int
            For a spatio-temporal searchlight:
                Two elements: the number of time-series and number of time
                samples for which patches are generated.
            For a spatial searchlight:
                One element: the number of time-series for which patches are
                generated.
            For a temporal searchlight:
                One element: the number of time-samples for which patches are
                generated.
            For no searchlight:
                Zero elements.
        """
        if (self.spatial_radius is not None
                and self.temporal_radius is not None):
            return (len(self.sel_series), len(self.time_centers))
        elif self.spatial_radius is not None:
            return (len(self.sel_series),)
        elif self.temporal_radius is not None:
            return (len(self.time_centers),)
        else:
            return tuple()

    def __len__(self):
        """Get total number of searchlight patches that will be generated."""
        total = 1
        for n in self.shape:  # Number of patches generated in each dimension
            total *= n
        return total


def _get_in_radius(dist, seed, radius):
    """Obtain indices for all points within the given radius from a seed point.

    Takes care to work with sparse matrices too.

    Parameters
    ----------
    dist : ndarray or sparse matrix, shape (n_points, n_points)
        The distances between all points.
    seed : int
        The index of the point used as a seed.
    radius : float
        The maximum distance that points can be to be included.

    Returns
    -------
    ind : ndarray, shape (n_points_in_radius,)
        Indices of all points in the given radius from the seed point.
    """
    from scipy.sparse import issparse
    if issparse(dist):
        # Treat all zero distances as missing data
        ind = dist[seed].nonzero()[1]
        # Find indices for points within the radius
        ind = ind[dist[seed].data < radius]
        ind.sort()
        # Be sure to add the seed point, which has distance of 0 to itself
        ind = np.hstack((ind, [seed]))  # Di
    else:
        ind = np.flatnonzero(dist[seed] < radius)
    return sorted(ind)



from functools import partial
import matplotlib.pyplot as plt
from mne.viz.topo import _iter_topography
import numpy as np
from scipy.spatial import distance


def plot_dsms(dsms, names=None, items=None, n_rows=1, cmap='viridis',
              title=None):
    """Plot one or more DSMs

    Parameters
    ----------
    dsms : ndarray | list of ndarray
        The DSM or list of DSMs to plot. The DSMs can either be two-dimensional
        (n_items x n_items) matrices or be in condensed form.
    names : str | list of str | None
        For each given DSM, a name to show above it. Defaults to no names.
    items : list of str | None
        The each item (row/col) in the DSM, a string description. This will be
        displayed along the axes. Defaults to None which means the items will
        be numbered.
    n_rows : int
        Number of rows to use when plotting multiple DSMs at once. Defaults to
        1.
    cmap : str
        Matplotlib colormap to use. See
        https://matplotlib.org/gallery/color/colormap_reference.html
        for all possibilities. Defaults to 'viridis'.
    title : str | None
        Title for the entire figure. Defaults to no title.

    Returns
    -------
    fig : matplotlib figure
        The figure produced by matplotlib
    """
    if not isinstance(dsms, list):
        dsms = [dsms]

    if isinstance(names, str):
        names = [names]
    if names is not None and len(names) != len(dsms):
        raise ValueError(f'Number of given names ({len(names)}) does not '
                         f'match the number of DSMs ({len(dsms)})')

    n_cols = int(np.ceil(len(dsms) / n_rows))
    fig = plt.figure(figsize=(2 * n_cols, 2 * n_rows))

    ax = fig.subplots(n_rows, n_cols, sharex=True, sharey=True, squeeze=False)
    for row in range(n_rows):
        for col in range(n_cols):
            i = row * n_cols + col % n_cols
            if i < len(dsms):
                dsm = dsms[i]
                if dsm.ndim == 1:
                    dsm = distance.squareform(dsm)
                elif dsm.ndim > 2:
                    raise ValueError(f'Invalid shape {dsm.shape} for DSM')
                im = ax[row, col].imshow(dsm, cmap=cmap)

                if names is not None:
                    name = names[i]
                    ax[row, col].set_title(name)
            else:
                ax[row, col].set_visible(False)

    plt.colorbar(im, ax=ax)
    if title is not None:
        plt.suptitle(title)
    return fig


def _click_func(ax, ch_idx, dsms, cmap):
    """ Function used to plot a single DSM interactively.

    Parameters
    ----------
    ax: matplotlib.Axes.axes
        Axes.axes object on which a new single DSM is plotted.
    ch_idx: int
        Index of a channel.
    dsms: ndarray, shape (n_sensors, n_dsm_datapoint)
        DSMs of MEG recordings; there's one DSM for each sensor.
    cmap: str
        Colormap used for plotting DSMs.
        Check matplotlib.pyplot.imshow for details.
    """
    dsm = dsms[ch_idx]
    dsm = distance.squareform(dsm)
    ax.imshow(dsm, cmap=cmap)


def _plot_dsms_topo_timepoint(dsms, info, layout=None, fig=None, title=None,
                              axis_facecolor='w', axis_spinecolor='w',
                              fig_facecolor='w', figsize=(6.4, 4.8),
                              cmap='viridis', show=False):
    """Plot DSMs on 2D MEG topography.

    Parameters
    ----------
    dsms: ndarray, shape (n_sensors, n_dsm_datapoints)
        DSMs of MEG recordings; one DSM for each sensor.
    info: mne.io.meas_info.Info
        Info object that contains meta data of MEG recordings.
    layout: mne.channels.layout.Layout | None
        Layout objects containing sensor layout info.
        The default (``None``) will figure out layout based on info.
    fig: matplotlib.pyplot.Figure | None
        Figure object on which DSMs on 2D MEG topography are plotted.
        The default (``None``) creates a new Figure object.
    title: str | None
        Title of the plot, used only when ``fig=None``.
        The default (``None``) puts no title in the figure.
    axis_facecolor: str
        Face color of the each DSM. Defaults to 'w', white.
    axis_spinecolor: str
        Spine color of each DSM. Defaults to 'w', white.
    fig_facecolor: str
        Face color of the entire topography. Defaults to 'w', white.
    figsize: tuple of float
        Figure size. The first element specify width and the second height.
        Defaults to (6.4, 4.8).
    cmap: str
        Colormap used for plotting DSMs. Defaults to 'viridis'.
        Check :func:`matplotlib.pyplot.imshow` for details.
    show: bool
        Whether to display the generated figure. Defaults to False.

    Returns
    -------
    fig: matplotlib.pyplot.Figure
        Figure object in which DSMs are plotted on 2D MEG topography.
    """
    on_pick = partial(_click_func, dsms=dsms, cmap=cmap)

    if fig is None:
        fig = plt.figure(figsize=figsize)
        if title is not None:
            fig.suptitle(title, x=0.98, horizontalalignment='right')
    else:
        fig = plt.figure(fig.number)

    my_topo_plot = _iter_topography(info=info, layout=layout, on_pick=on_pick,
                                    fig=fig, axis_facecolor=axis_facecolor,
                                    axis_spinecolor=axis_spinecolor,
                                    fig_facecolor=fig_facecolor,
                                    unified=False)

    for i, (ax, _) in enumerate(my_topo_plot):
        dsms_i = dsms[i]
        dsms_i = distance.squareform(dsms_i)
        ax.imshow(dsms_i, cmap=cmap)

    if show:
        fig.show()

    return fig


def plot_dsms_topo(dsms, info, time=None, layout=None, fig=None,
                   axis_facecolor='w', axis_spinecolor='w', fig_facecolor='w',
                   figsize=(6.4, 4.8), cmap='viridis', show=True):
    """ Plot DSMs on 2D sensor topography

    Parameters
    ----------
    dsms: ndarray | numpy.memmap, shape (n_sensors,[ n_times,] n_dsm_datapts)
        DSMs of MEG/EEG recordings; one DSM for each sensor and time point.
    info: mne.io.meas_info.Info
        Info object that contains meta data of MEG/EEG recordings.
    time: int | [int, int] | None
        A time point (int) or time window ([int, int]) for which DSMs are
        plotted. When a time window is given, averge DSMs for the window are
        plotted. The default (``None``) plots the average DSMs of all the time
        points. Start of the time window is inclusive, while the end is
        exclusive.
    layout: mne.channels.layout.Layout, optional
        Layout objects containing sensor layout info.
        The default, ``layout=None``, will figure out layout based on info.
    fig: matplotlib.pyplot.Figure | None, optional
        Figure object on which DSMs on 2D sensor topography are plotted.
        The default (``None``) creates a new Figure object
        with a title based on time parameter.
    axis_facecolor: str, optional
        Face color of the each DSM. Defaults to 'w', white.
    axis_spinecolor: str, optional
        Spine color of each DSM. Defaults to 'w', white.
    fig_facecolor: str, optional
        Face color of the entire topography. Defaults to 'w', white.
    figsize: tuple of float, optional
        Figure size. The first element specify width and the second height.
        Defaults to (6.4, 4.8).
    cmap: str, optional
        Colormap used for plotting DSMs. Defaults to 'viridis'.
        Check :func:`matplotlib.pyplot.imshow` for details.
    show: bool, optional
        Whether to display the generated figure. Defaults to ``True``.

    Returns
    -------
    fig: matplotlib.pyplot.Figure
        Figure object in which DSMs are plotted on 2D sensor topography.
    """
    if dsms.ndim != 2 and dsms.ndim != 3:
        raise ValueError('dsms have to be a 2D or 3D ndarray or numpy.memmap, '
                         '[n_sensors,[ n_times,] n_dsm_datapoints]')
    if len(dsms.shape) == 2:
        dsms = dsms[:, np.newaxis, :]
    if time is None:
        time = [0, dsms.shape[1]]
    if isinstance(time, int):
        time = [time, time + 1]
    if not isinstance(time, list):
        raise TypeError('time has to be int, list of [int, int] or None.')
    if (not all(isinstance(i, int) for i in time)) or (len(time) != 2):
        raise TypeError('time has to be int, list of [int, int] or None.')
    if time[0] >= time[1]:
        raise ValueError('The start of the time window has to be smaller '
                         'than the end of the time window.')
    if time[0] < 0 or time[1] > dsms.shape[1]:
        raise ValueError('The time window is out of range. The minimum is 0 '
                         f'and the maximum is {dsms.shape[1]}')
    if (fig is not None) and (not isinstance(fig, plt.Figure)):
        raise TypeError('fig has to be matplotlib.pyplot.Figure or None.')

    dsms_cropped = dsms[:, time[0]:time[1], :]
    dsms_avg = dsms_cropped.mean(axis=1)
    # set title to time window
    if time[0] + 1 != time[1]:
        title = f'From {time[0]} (inclusive) to {time[1]} (exclusive)'
    else:
        title = f'Time point: {time[0]}'

    fig = _plot_dsms_topo_timepoint(dsms_avg, info, fig=fig, layout=layout,
                                    title=title,
                                    axis_facecolor=axis_facecolor,
                                    axis_spinecolor=axis_spinecolor,
                                    fig_facecolor=fig_facecolor,
                                    figsize=figsize, cmap=cmap, show=show)
    return fig



# encoding: utf-8
"""
Methods to compute dissimilarity matrices (DSMs).
"""

import numpy as np
from scipy.spatial import distance
from mne.utils import logger

from .folds import _create_folds
from .searchlight import searchlight


def compute_dsm(data, metric='correlation', **kwargs):
    """Compute a dissimilarity matrix (DSM).

    Parameters
    ----------
    data : ndarray, shape (n_items, ...)
        For each item, all the features. The first are the items and all other
        dimensions will be flattened and treated as features.
    metric : str | function
        The distance metric to use to compute the DSM. Can be any metric
        supported by :func:`scipy.spatial.distance.pdist`. When a function is
        specified, it needs to take in two vectors and output a single number.
        See also the ``dist_params`` parameter to specify and additional
        parameter for the distance function.
        Defaults to 'correlation'.
    **kwargs : dict, optional
        Extra arguments for the distance metric. Refer to
        :mod:`scipy.spatial.distance` for a list of all other metrics and their
        arguments.

    Returns
    -------
    dsm : ndarray, shape (n_classes * n_classes-1,)
        The DSM, in condensed form.
        See :func:`scipy.spatial.distance.squareform`.

    See Also
    --------
    compute_dsm_cv
    """
    X = np.reshape(np.asarray(data), (len(data), -1))
    n_items, n_features = X.shape

    # Be careful with certain metrics
    if n_features == 1 and metric in ['correlation', 'cosine']:
        raise ValueError("There is only a single feature, so "
                         "'correlataion' and 'cosine' can not be "
                         "used as DSM metric. Consider using 'sqeuclidean' "
                         "instead.")

    return distance.pdist(X, metric, **kwargs)


def compute_dsm_cv(folds, metric='correlation', **kwargs):
    """Compute a dissimilarity matrix (DSM) using cross-validation.

    The distance computation is performed from the average of
    all-but-one "training" folds to the remaining "test" fold. This is repeated
    with each fold acting as the "test" fold once. The resulting distances are
    averaged and the result used in the final DSM.

    Parameters
    ----------
    folds : ndarray, shape (n_folds, n_items, ...)
        For each item, all the features. The first dimension are the folds used
        for cross-validation, items are along the second dimension, and all
        other dimensions will be flattened and treated as features.
    metric : str | function
        The distance metric to use to compute the DSM. Can be any metric
        supported by :func:`scipy.spatial.distance.pdist`. When a function is
        specified, it needs to take in two vectors and output a single number.
        See also the ``dist_params`` parameter to specify and additional
        parameter for the distance function.
        Defaults to 'correlation'.
    **kwargs : dict, optional
        Extra arguments for the distance metric. Refer to
        :mod:`scipy.spatial.distance` for a list of all other metrics and their
        arguments.

    Returns
    -------
    dsm : ndarray, shape (n_classes * n_classes-1,)
        The cross-validated DSM, in condensed form.
        See :func:`scipy.spatial.distance.squareform`.

    See Also
    --------
    compute_dsm
    """
    X = np.reshape(folds, (folds.shape[0], folds.shape[1], -1))
    n_folds, n_items, n_features = X.shape[:3]

    # Be careful with certain metrics
    if n_features == 1 and metric in ['correlation', 'cosine']:
        raise ValueError("There is only a single feature, so "
                         "'correlataion' and 'cosine' can not be "
                         "used as DSM metric. Consider using 'sqeuclidean' "
                         "instead.")

    dsm = np.zeros((n_items * (n_items - 1)) // 2)

    X_mean = X.mean(axis=0)

    # Do cross-validation
    for test_fold in range(n_folds):
        X_test = X[test_fold]
        X_train = X_mean - (X_mean - X_test) / (n_folds - 1)

        dist = distance.cdist(X_train, X_test, metric, **kwargs)
        dsm += dist[np.triu_indices_from(dist, 1)]

    return dsm / n_folds


def _ensure_condensed(dsm, var_name):
    """Converts a DSM to condensed form if needed."""
    if type(dsm) is list:
        return [_ensure_condensed(d, var_name) for d in dsm]

    if not isinstance(dsm, np.ndarray):
        raise TypeError('A single DSM should be a NumPy array. '
                        'Multiple DSMs should be a list of NumPy arrays.')

    if dsm.ndim == 2:
        if dsm.shape[0] != dsm.shape[1]:
            raise ValueError(f'Invalid dimensions for "{var_name}" '
                             '({dsm.shape}). The DSM should either be a '
                             'square matrix, or a one dimensional array when '
                             'in condensed form.')
        dsm = distance.squareform(dsm)
    elif dsm.ndim != 1:
        raise ValueError(f'Invalid dimensions for "{var_name}" ({dsm.shape}). '
                         'The DSM should either be a square matrix, or a one '
                         'dimensional array when in condensed form.')
    return dsm


def _n_items_from_dsm(dsm):
    """Get the number of items, given a DSM."""
    if dsm.ndim == 2:
        return dsm.shape[0]
    elif dsm.ndim == 1:
        return distance.squareform(dsm).shape[0]


class dsm_array:
    """Generate DSMs from an array of data, possibly in a searchlight pattern.

    First use :class:`searchlight_patches` to compute the searchlight patches.
    Then you can use this function to compute DSMs for each searchlight patch.

    Parameters
    ----------
    X : ndarray, shape (n_items, n_series, n_times)
        An array containing the data.
    patches : generator of tuples | None
        Searchlight patches as generated by :class:`searchlight`. If ``None``,
        no searchlight is used. Defaults to ``None``.
    dist_metric : str | function
        The distance metric to use to compute the DSMs. Can be any metric
        supported by :func:`scipy.spatial.distance.pdist`. When a function is
        specified, it needs to take in two vectors and output a single number.
        See also the ``dist_params`` parameter to specify and additional
        parameter for the distance function.
        Defaults to 'correlation'.
    dist_params : dict
        Extra arguments for the distance metric used to compute the DSMs.
        Refer to :mod:`scipy.spatial.distance` for a list of all other metrics
        and their arguments. Defaults to an empty dictionary.
    y : ndarray of int, shape (n_items,) | None
        For each item, a number indicating the class to which the item belongs.
        When ``None``, each item is assumed to belong to a different class.
        Defaults to ``None``.
    n_folds : int | None
        Number of cross-validation folds to use when computing the distance
        metric. Folds are created based on the ``y`` parameter. Specify
        ``None`` to use the maximum number of folds possible, given the data.
        Defaults to 1 (no cross-validation).

    Yields
    ------
    dsm : ndarray, shape (n_patches, n_items * (n_items - 1))
        A DSM (in condensed form) for each searchlight patch.

    Attributes
    ----------
    shape : tuple of int
        Multidimensional shape of the generted DSMs.

        This is useful for re-shaping the result obtained after consuming the
        this generator.

        For a spatio-temporal searchlight:
            Three elements: the number of time-series, number of time
            samples and length of a consensed DSM.
        For a spatial searchlight:
            Two element: the number of time-series and length of a condensed
            DSM.
        For a temporal searchlight:
            Two elements: the number of time-samples and length of a condensed
            DSM.
        For no searchlight:
            One element: the length of a condensed DSM.

    See also
    --------
    dsm
    rsa
    searchlight
    """
    def __init__(self, X, patches=None, dist_metric='correlation',
                 dist_params=dict(), y=None, n_folds=1, sel_series=None,
                 sel_samples=None):
        if patches is None:
            patches = searchlight(X.shape)

        # Create folds for cross-validated DSM metrics
        self.X = _create_folds(X, y, n_folds)
        # The data is now folds x items x n_series x ...

        self.patches = patches
        self.dist_metric = dist_metric
        self.dist_params = dist_params
        self.use_cv = len(self.X) > 1  # More than one fold present

        # Target shape for an array that would hold all of the generated DSMs.
        dsm_length = len(np.triu_indices(self.X.shape[1], k=1)[0])
        self.shape = patches.shape + (dsm_length,)

        # Setup the generator that will be producing the DSMs
        self._generator = self._iter_dsms()

    def __iter__(self):
        return self

    def __next__(self):
        """Generate a DSM for each searchlight patch."""
        return next(self._generator)

    def _iter_dsms(self):
        for patch in self.patches:
            if self.use_cv:
                yield compute_dsm_cv(self.X[(slice(None),) + patch],
                                     self.dist_metric, **self.dist_params)
            else:
                yield compute_dsm(self.X[0][patch],
                                  self.dist_metric, **self.dist_params)

    def __len__(self):
        """Get total number of DSMs that will be generated."""
        return len(self.patches)



# encoding: utf-8
"""
Methods to compute representational similarity analysis (RSA).
"""

import numpy as np
from scipy import stats
from scipy.stats.stats import _kendall_dis
from joblib import Parallel, delayed

from .folds import _create_folds
from .dsm import _ensure_condensed, compute_dsm, compute_dsm_cv
from .searchlight import searchlight


def _kendall_tau_a(x, y):
    """Compute Kendall's Tau metric, A-variant.

    Taken from scipy.stats.kendalltau and modified to be the tau-a variant.
    """
    x = np.asarray(x).ravel()
    y = np.asarray(y).ravel()

    if x.size != y.size:
        raise ValueError("All inputs to `kendalltau` must be of the same size,"
                         " found x-size %s and y-size %s" % (x.size, y.size))
    elif not x.size or not y.size:
        return np.nan  # Return NaN if arrays are empty

    def count_rank_tie(ranks):
        cnt = np.bincount(ranks).astype('int64', copy=False)
        cnt = cnt[cnt > 1]
        return ((cnt * (cnt - 1) // 2).sum(),
                (cnt * (cnt - 1.) * (cnt - 2)).sum(),
                (cnt * (cnt - 1.) * (2 * cnt + 5)).sum())

    size = x.size
    perm = np.argsort(y)  # sort on y and convert y to dense ranks
    x, y = x[perm], y[perm]
    y = np.r_[True, y[1:] != y[:-1]].cumsum(dtype=np.intp)

    # stable sort on x and convert x to dense ranks
    perm = np.argsort(x, kind='mergesort')
    x, y = x[perm], y[perm]
    x = np.r_[True, x[1:] != x[:-1]].cumsum(dtype=np.intp)

    dis = _kendall_dis(x, y)  # discordant pairs

    obs = np.r_[True, (x[1:] != x[:-1]) | (y[1:] != y[:-1]), True]
    cnt = np.diff(np.nonzero(obs)[0]).astype('int64', copy=False)

    ntie = (cnt * (cnt - 1) // 2).sum()  # joint ties
    xtie, x0, x1 = count_rank_tie(x)     # ties in x, stats
    ytie, y0, y1 = count_rank_tie(y)     # ties in y, stats

    tot = (size * (size - 1)) // 2

    if xtie == tot or ytie == tot:
        return np.nan

    # Note that tot = con + dis + (xtie - ntie) + (ytie - ntie) + ntie
    #               = con + dis + xtie + ytie - ntie
    con_minus_dis = tot - xtie - ytie + ntie - 2 * dis
    tau = con_minus_dis / tot
    # Limit range to fix computational errors
    tau = min(1., max(-1., tau))

    return tau


def _partial_correlation(dsm_data, dsm_model, type='pearson'):
    """Compute partial Pearson/Spearman correlation."""
    if len(dsm_model) == 1:
        raise ValueError('Need more than one model DSM to use partial '
                         'correlation as metric.')
    if type not in ['pearson', 'spearman']:
        raise ValueError("Correlation type must by either 'pearson' or "
                         "'spearman'")
    X = np.vstack([dsm_data] + dsm_model).T
    if type == 'spearman':
        X = np.apply_along_axis(stats.rankdata, 0, X)
    X = X - X.mean(axis=0)
    cov_X_inv = np.linalg.pinv(X.T @ X)
    norm = np.sqrt(np.outer(np.diag(cov_X_inv), np.diag(cov_X_inv)))
    R_partial = cov_X_inv / norm
    return -R_partial[0, 1:]


def rsa_gen(dsm_data_gen, dsm_model, metric='spearman'):
    """Generate RSA values between data and model DSMs.

    Will yield RSA scores for each data DSM.

    Parameters
    ----------
    dsm_data_gen : generator of ndarray, shape (n_items, n_items)
        The generator for data DSMs
    dsm_model : ndarray, shape (n_items, n_items) | list of ndarray
        The model DSM, or list of model DSMs.
    metric : str
        The RSA metric to use to compare the DSMs. Valid options are:

        * 'spearman' for Spearman's correlation (the default)
        * 'pearson' for Pearson's correlation
        * 'kendall-tau-a' for Kendall's Tau (alpha variant)
        * 'partial' for partial Pearson correlations
        * 'partial-spearman' for partial Spearman correlations
        * 'regression' for linear regression weights

        Defaults to 'spearman'.

    Yields
    -------
    rsa_val : float | ndarray, shape (len(dsm_model),)
        For each data DSM, the representational similarity with the model DSM.
        When multiple model DSMs are specified, this will be a 1D array of
        similarities, comparing the data DSM with each model DSM.

    See also
    --------
    rsa
    """
    if type(dsm_model) == list:
        return_array = True
        dsm_model = [_ensure_condensed(dsm, 'dsm_model') for dsm in dsm_model]
    else:
        return_array = False
        dsm_model = [_ensure_condensed(dsm_model, 'dsm_model')]

    for dsm_data in dsm_data_gen:
        dsm_data = _ensure_condensed(dsm_data, 'dsm_data')
        rsa_vals = _rsa_single_dsm(dsm_data, dsm_model, metric)
        if return_array:
            yield np.asarray(rsa_vals)
        else:
            yield rsa_vals[0]


def _rsa_single_dsm(dsm_data, dsm_model, metric):
    """Compute RSA between a single data DSM and one or more model DSMs."""
    if metric == 'spearman':
        rsa_vals = [stats.spearmanr(dsm_data, dsm_model_)[0]
                    for dsm_model_ in dsm_model]
    elif metric == 'pearson':
        rsa_vals = [stats.pearsonr(dsm_data, dsm_model_)[0]
                    for dsm_model_ in dsm_model]
    elif metric == 'kendall-tau-a':
        rsa_vals = [_kendall_tau_a(dsm_data, dsm_model_)
                    for dsm_model_ in dsm_model]
    elif metric == 'partial':
        rsa_vals = _partial_correlation(dsm_data, dsm_model)
    elif metric == 'partial-spearman':
        rsa_vals = _partial_correlation(dsm_data, dsm_model,
                                        type='spearman')
    elif metric == 'regression':
        X = np.atleast_2d(np.array(dsm_model)).T
        X = X - X.mean(axis=0)
        y = dsm_data - dsm_data.mean()
        rsa_vals = np.linalg.lstsq(X, y, rcond=None)[0]
    else:
        raise ValueError("Invalid RSA metric, must be one of: 'spearman', "
                         "'pearson', 'partial', 'partial-spearman', "
                         "'regression' or 'kendall-tau-a'.")
    return rsa_vals


def rsa(dsm_data, dsm_model, metric='spearman', n_data_dsms=None, n_jobs=1,
        verbose=False):
    """Perform RSA between data and model DSMs.

    Parameters
    ----------
    dsm_data : ndarray, shape (n_items, n_items) | list | generator
        The data DSM (or list/generator of data DSMs).
    dsm_model : ndarray, shape (n_items, n_items) | list of ndarray
        The model DSM (or list of model DSMs).
    metric : str
        The RSA metric to use to compare the DSMs. Valid options are:

        * 'spearman' for Spearman's correlation (the default)
        * 'pearson' for Pearson's correlation
        * 'kendall-tau-a' for Kendall's Tau (alpha variant)
        * 'partial' for partial Pearson correlations
        * 'partial-spearman' for partial Spearman correlations
        * 'regression' for linear regression weights

        Defaults to 'spearman'.

    n_data_dsms : int | None
        The number of data DSMs. This is useful when displaying a progress bar,
        so an estimate can be made of the computation time remaining. This
        information is available if ``dsm_data`` is an array or a list, but if
        it is a generator, this information is not available and you may want
        to set it explicitly.
    n_jobs : int
        The number of processes (=number of CPU cores) to use. Specify -1 to
        use all available cores. Defaults to 1.
    verbose : bool
        Whether to display a progress bar. In order for this to work, you need
        the tqdm python module installed. Defaults to False.

    Returns
    -------
    rsa_val : float | ndarray, shape (len(dsm_data), len(dsm_model))
        Depending on whether one or more data and model DSMs were specified, a
        single similarity value or a 2D array of similarity values for each
        data DSM versus each model DSM.

    See also
    --------
    rsa_gen
    """
    return_array = False
    if type(dsm_data) == list or hasattr(dsm_data, '__next__'):
        return_array = True
    else:
        dsm_data = [dsm_data]

    if verbose:
        from tqdm import tqdm
        if n_data_dsms is not None:
            total = n_data_dsms
        elif hasattr(dsm_data, '__len__'):
            total = len(dsm_data)
        else:
            total = None
        dsm_data = tqdm(dsm_data, total=total, unit='DSM')

    if n_jobs == 1:
        rsa_vals = list(rsa_gen(dsm_data, dsm_model, metric))
    else:
        def process_single_dsm(dsm):
            return next(rsa_gen([dsm], dsm_model, metric))
        rsa_vals = Parallel(n_jobs)(delayed(process_single_dsm)(dsm)
                                    for dsm in dsm_data)
    if return_array:
        return np.asarray(rsa_vals)
    else:
        return rsa_vals[0]


def rsa_array(X, dsm_model, patches=None, data_dsm_metric='correlation',
              data_dsm_params=dict(), rsa_metric='spearman', y=None, n_folds=1,
              n_jobs=1, verbose=False):
    """Perform RSA on an array of data, possibly in a searchlight pattern.

    Parameters
    ----------
    X : ndarray, shape (n_items, n_series, n_times)
        An array containing the data.
    dsm_model : ndarray, shape (n, n) | (n * (n - 1) // 2,) | list of ndarray
        The model DSM, see :func:`compute_dsm`. For efficiency, you can give it
        in condensed form, meaning only the upper triangle of the matrix as a
        vector. See :func:`scipy.spatial.distance.squareform`. To perform RSA
        against multiple models at the same time, supply a list of model DSMs.

        Use :func:`compute_dsm` to compute DSMs.
    patches : generator of tuples | None
        Searchlight patches as generated by :class:`searchlight`. If ``None``,
        no searchlight is used. Defaults to ``None``.
    data_dsm_metric : str
        The metric to use to compute the data DSMs. This can be any metric
        supported by the scipy.distance.pdist function. Defaults to
        'correlation'.
    data_dsm_params : dict
        Extra arguments for the distance metric used to compute the DSMs.
        Refer to :mod:`scipy.spatial.distance` for a list of all other metrics
        and their arguments. Defaults to an empty dictionary.
    rsa_metric : str
        The RSA metric to use to compare the DSMs. Valid options are:

        * 'spearman' for Spearman's correlation (the default)
        * 'pearson' for Pearson's correlation
        * 'kendall-tau-a' for Kendall's Tau (alpha variant)
        * 'partial' for partial Pearson correlations
        * 'partial-spearman' for partial Spearman correlations
        * 'regression' for linear regression weights

        Defaults to 'spearman'.
    y : ndarray of int, shape (n_items,) | None
        For each item, a number indicating the class to which the item belongs.
        When ``None``, each item is assumed to belong to a different class.
        Defaults to ``None``.
    n_folds : int | None
        Number of cross-validation folds to use when computing the distance
        metric. Folds are created based on the ``y`` parameter. Specify -1 to
        use the maximum number of folds possible, given the data.
        Defaults to 1 (no cross-validation).
    n_jobs : int
        The number of processes (=number of CPU cores) to use. Specify -1 to
        use all available cores. Defaults to 1.
    verbose : bool
        Whether to display a progress bar. In order for this to work, you need
        the tqdm python module installed. Defaults to False.

    Returns
    -------
    rsa_vals : ndarray, shape ([n_series,] [n_times,] [n_model_dsms])
        The RSA value for each searchlight patch. When ``spatial_radius`` is
        set to ``None``, there will only be no ``n_series`` dimension. When
        ``temporal_radius`` is set to ``None``, there will be no time
        dimension. When multiple models have been supplied, the last dimension
        will contain RSA results for each model.

    See Also
    --------
    searchlight
    compute_dsm
    dsm_array
    """
    if patches is None:
        patches = searchlight(X.shape)  # One big searchlight patch

    # Create folds for cross-validated DSM metrics
    X = _create_folds(X, y, n_folds)
    # The data is now folds x items x n_series x n_times

    if type(dsm_model) == list:
        dsm_model = [_ensure_condensed(dsm, 'dsm_model') for dsm in dsm_model]
    else:
        dsm_model = [_ensure_condensed(dsm_model, 'dsm_model')]

    if verbose:
        from tqdm import tqdm
        shape = patches.shape
        patches = tqdm(patches, unit='patch')
        patches.shape = shape

    def rsa_single_patch(patch):
        """Compute RSA for a single searchlight patch."""
        if len(X) == 1:  # Check number of folds
            # No cross-validation
            dsm_data = compute_dsm(X[0][patch],
                                   data_dsm_metric, **data_dsm_params)
        else:
            # Use cross-validation
            dsm_data = compute_dsm_cv(X[(slice(None),) + patch],
                                      data_dsm_metric, **data_dsm_params)
        return _rsa_single_dsm(dsm_data, dsm_model, metric=rsa_metric)

    # Call RSA multiple times in parallel for each searchlight patch
    data = Parallel(n_jobs)(delayed(rsa_single_patch)(patch)
                            for patch in patches)

    # Figure out the desired dimensions of the resulting array
    dims = patches.shape
    if len(dsm_model) > 1:
        dims = dims + (len(dsm_model),)

    return np.array(data).reshape(dims)



__version__ = '0.4'
from .source_level import rsa_stcs, dsm_stcs, rsa_nifti, dsm_nifti
from .sensor_level import rsa_evokeds, rsa_epochs, dsm_evokeds, dsm_epochs
from .searchlight import searchlight
from .rsa import rsa, rsa_gen, rsa_array
from .dsm import compute_dsm, compute_dsm_cv, dsm_array
from .viz import plot_dsms, plot_dsms_topo



#! /usr/bin/env python
from setuptools import setup
import os
import os.path as op

# get the version (don't import mne here, so dependencies are not needed)
version = None
with open(op.join('mne_rsa', '__init__.py'), 'r') as fid:
    for line in (line.strip() for line in fid):
        if line.startswith('__version__'):
            version = line.split('=')[1].strip().strip('\'')
            break
if version is None:
    raise RuntimeError('Could not determine version')

if __name__ == "__main__":
    if os.path.exists('MANIFEST'):
        os.remove('MANIFEST')

    with open('README.rst', 'r') as fid:
        long_description = fid.read()

    setup(
        name='mne-rsa',
        maintainer='Marijn van Vliet',
        maintainer_email='w.m.vanvliet@gmail.com',
        description='Code for performing Representational Similarity Analysis on MNE-Python data structures.',
        license='BSD-3',
        url='https://github.com/wmvanvliet/mne-rsa',
        version=version,
        download_url='https://github.com/wmvanvliet/mne-rsa/archive/master.zip',
        long_description=long_description,
        classifiers=['Intended Audience :: Science/Research',
                     'Intended Audience :: Developers',
                     'License :: OSI Approved',
                     'Programming Language :: Python',
                     'Topic :: Software Development',
                     'Topic :: Scientific/Engineering',
                     'Operating System :: Microsoft :: Windows',
                     'Operating System :: POSIX',
                     'Operating System :: Unix',
                     'Operating System :: MacOS'],
        platforms='any',
        packages=['mne_rsa'],
        install_requires=['numpy', 'scipy', 'matplotlib', 'mne'],
    )
